from mext.threads.locks import SimpleLock, TimeoutLock
from mext.classhacks import get_result
from thread import get_ident

__all__ = ['Event', 'Flag', 'DeferredResult', 'DeferredCall', 'Resource']


class Event(object):
    """
        A primitive to synchronize changes and threads checing / waiting for those
        changes.

            with evt:
                while not expected state:
                    evt.wait() # wait for relevant state changes


            with evt:
                change state
                evt.notify(1)
    """
    def __init__(self, lock=None):
        self.lock = lock or SimpleLock()
        self.handlers = []

    def __enter__(self):
        return self.lock.__enter__()

    def __exit__(self, *args):
        return self.lock.__exit__(*args)

    def __repr__(self):
        return "<%s(%d)>" % (self.__class__.__name__, len(self.handlers))

    def _check_locked(self):
        assert self.lock.locked(), "cannot wait on un-acquired lock"

    def wait(self, timeout=None):
        self._check_locked()
        if not timeout: # None or 0 can be served by simple Lock
            hlock = SimpleLock()
        else:
            hlock = TimeoutLock()
        hlock.acquire()
        self.handlers.append(hlock)
        self.lock.release()
        try:
            if not timeout:
                hlock.acquire(timeout is None)
                return True
            else:
                r = hlock.acquire(timeout)
        finally:
            self.lock.acquire()
        if not r:
            try:
                self.handlers.remove(hlock)
            except ValueError:
                # this happens if this handler was notified after it timed out
                # but before we got self.lock, so it's the same as if we got it
                assert timeout
                return True
        return r

    def notify(self, n): # pass None to notify all
        self._check_locked()
        # order of self.handlers operations does not matter as they
        # always happen with a lock
        for hlock in self.handlers[:n]:
            hlock.release()
        self.handlers[:n] = []







class Flag(object):
    def __init__(self):
        self.evt_set = Event()
        self.flag = False

    def __nonzero__(self):
        return self.flag

    def set(self):
        with self.evt_set:
            assert not self.flag, "Flag is already set"
            self.flag = True
            self.evt_set.notify(None)

    def clear(self):
        with self.evt_set:
            self.flag = False

    def wait(self, timeout=None):
        with self.evt_set:
            if not self.flag:
                self.evt_set.wait(timeout)


_not_set = object()

class DeferredResult(object):
    def __init__(self):
        self.ready = Event()
        self.result = _not_set

    def set(self, value):
        with self.ready:
            assert self.result is _not_set, "Value is already set"
            self.result = value
            self.ready.notify(None)

    def get(self):
        # avoid locking if possible
        if self.result is _not_set:
            self.wait()
        return self.result

    def wait(self, timeout=None):
        with self.ready:
            # if called from .get we check again,
            # because first check was outside the lock
            if self.result is _not_set:
                return self.ready.wait(timeout)

class DeferredCall(DeferredResult):
    def __init__(self, *args, **kw):
        DeferredResult.__init__(self)
        self.args = args
        self.kw = kw

    def execute(self):
        self.set(get_result(*self.args, **self.kw))

    def get(self):
        return DeferredResult.get(self)()




class Resource(object):
    def __init__(self, count=1, verbose=None):
        if count < 0:
            raise ValueError("Initial resource count must be >= 0")
        self.evt_put = Event()
        self.count = count

    def get(self, blocking=True):
        with self.evt_put:
            while not self.count:
                if not blocking:
                    return False
                self.evt_put.wait()
            self.count -= 1
            return True

    __enter__ = get

    def put(self, count=1):
        with self.evt_put:
            self.count += count
            self.evt_put.notify(count)

    def __exit__(self, t, v, tb):
        self.put()
